from __future__ import annotations

from itertools import product, repeat
from typing import Any, Union, Optional, Tuple, Mapping, Literal

import aiger
import aiger_bv as BV
from bidict import bidict
from dfa import DFA
from pyrsistent import pmap
from pyrsistent.typing import PMap


Kinds = Union[Literal['inputs'], Literal['outputs'], Literal['states']]
Encoding = Mapping[str, bidict]  # Bijective Mapping[str, bitvector]]
Encodings = Mapping[Kinds, Encoding]
State = Tuple[PMap[str, Any], PMap[str, Any]]  # (prev_output, latch_val).


def default_encoding(bmap, is_output=False):
    enc = {}
    flatten = (zip(repeat(k), range(1 << size)) for k, size in bmap.items())
    for elems in product(*flatten):
        decoded = pmap({
            k: tuple(BV.encode_int(bmap[k].size, v, signed=False))
            for k, v in elems
        })
        enc[decoded] = decoded
    return enc


def aig2dfa(
        circ: BV.AIGBV,
        relabels: Encodings = pmap(),
        initial_label: Optional[Any] = None,
) -> DFA:
    """
    Converts an AIG circuit into a dfa.DFA object.

    Note that circuits are generally Mealy Machines
       - i.e., label depends on input & state.

    If the initial output is not included, the initial label is None.

    Relabels is an optional mapping with 'inputs' and 'outputs' keys.
     - 'inputs' indexes a mapping from dfa inputs to (pmap) aiger inputs.
     - 'outputs' indexes a mapping from (pmap) aiger outputs to dfa outputs.
     - pmap = pyrsistent.pmap

    Returns:
      A returned DFA is the Moore machine encoding of this Mealy
      Machine. This means that the state includes the previous output.
    """
    if isinstance(circ, aiger.AIG):
        circ = BV.aig2aigbv(circ)

    if 'inputs' not in relabels:
        relabels = relabels.set('inputs', default_encoding(circ.imap))
    if 'outputs' not in relabels:
        relabels = relabels.set('outputs', default_encoding(circ.omap))

    def transition(state, action):
        action = relabels['inputs'][action]
        omap, lmap = circ(action, latches=state[1])
        output = relabels['outputs'][pmap(omap)]
        return output, pmap(lmap)

    outputs = set(relabels['outputs'].values())
    if initial_label is None:
        outputs.add(None)

    return DFA(
        start=(initial_label, pmap(circ.latch2init)),
        inputs=relabels['inputs'].keys(),
        outputs=outputs,
        label=lambda s: s[0],
        transition=transition,
    )


__all__ = ["aig2dfa", "State"]



from dfa import DFA
from pyrsistent import pmap

from aiger_dfa import dfa2aig


def test_dfa2aig():
    dfa1 = DFA(
        start=0,
        inputs={0, 1},
        label=lambda s: (s % 4) == 3,
        transition=lambda s, c: (s + c) % 4,
    )
    circ, relabels, valid = dfa2aig(dfa1)
    simulator = circ.simulator()
    vsimulator = valid.simulator()
    next(simulator)
    next(vsimulator)

    in2bv = relabels['inputs']

    def step(action):
        action_bv = in2bv[action]

        vout, _ = vsimulator.send(action_bv)
        is_valid = vout['valid'][0]

        out, lout = simulator.send(action_bv)
        label = relabels['outputs'][pmap(out)]
        return label, is_valid, lout['state']

    prev_state = circ.latch2init['state']
    for i in range(5):
        label, valid, state = step(0)
        assert valid
        assert state == prev_state
        assert dfa1.label((0,)*i) == label

    for i in range(5):
        label, valid, state = step(1)
        assert valid
        assert dfa1.label((1,)*i) == label

    # One invalid input should invalidate the run.

    vout, _ = vsimulator.send({'action': (True, True)})
    assert not vout['valid'][0]

    for i in range(5):
        _, valid, _ = step(0)
        assert not valid



from dfa import DFA, dict2dfa

from aiger_dfa import dfa2aig, aig2dfa


def test_dfa2aig():
    dfa1 = DFA(
        start=0,
        inputs={0, 1},
        label=lambda s: (s % 4) == 3,
        transition=lambda s, c: (s + c) % 4,
    )
    circ, relabels, _ = dfa2aig(dfa1)
    dfa2 = aig2dfa(circ, relabels)
    for i in range(1, 5):
        word = (0,)*i
        assert dfa1.label(word) == dfa2.label(word)

        # Moore encoding of Mealy. Need 1 step delay.
        word = (1,) * i
        assert dfa1.label(word) == dfa2.label(word + (0,))


def test_dfa2aig_default_encoding():
    dfa1 = dict2dfa({
        'foo': (False, {'a': 'foo', 'b': 'bar'}),
        'bar': (True, {'a': 'foo', 'b': 'foo'})
    }, 'foo')
    circ1, *_ = dfa2aig(dfa1)
    dfa2 = aig2dfa(circ1)
    assert len(dfa2.inputs) == 4
    assert len(dfa2.outputs) == 4 + 1  # Moore Encoding.


def test_dfa2aig_smoke():
    dfa1 = dict2dfa({
        'foo': (False, {'a': 'foo', 'b': 'bar'}),
        'bar': (True, {'a': 'foo', 'b': 'foo'})
    }, 'foo')
    circ1, relabels, _ = dfa2aig(dfa1)
    dfa2 = aig2dfa(circ1, relabels)

    assert dfa2.inputs == dfa1.inputs
    assert dfa2.outputs == dfa1.outputs | {None}

    word = ('a', 'b', 'a', 'a', 'b')
    assert dfa1.label(word) == dfa2.label(word + ('a',))


def test_dfa2aig_smoke_lift_aig():
    dfa1 = dict2dfa({
        'foo': (False, {'a': 'foo', 'b': 'bar'}),
        'bar': (True, {'a': 'foo', 'b': 'foo'})
    }, 'foo')
    circ1, *_ = dfa2aig(dfa1)
    dfa2 = aig2dfa(circ1.aig)
    len(dfa1.states()) == len(dfa2.states())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger_dfa']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.0.0,<22.0.0',
 'bidict>=0.21.0,<0.22.0',
 'dfa>=2.0.0,<3.0.0',
 'funcy>=1.12,<2.0',
 'py-aiger-bv>=4.0.0,<5.0.0',
 'py-aiger-ptltl>=3.0.0,<4.0.0',
 'py-aiger>=6.0.0,<7.0.0',
 'pyrsistent>=0.18,<0.19']

setup_kwargs = {
    'name': 'py-aiger-dfa',
    'version': '0.3.2',
    'description': 'Library for moving between sequential circuits AIGs and DFAs.',
    'long_description': "# py-aiger-dfa\nPython library for converting between AIG circuits and DFAs.\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger-dfa/status.svg)](https://cloud.drone.io/mvcisback/py-aiger-dfa)\n[![Docs](https://img.shields.io/badge/API-link-color)](https://mvcisback.github.io/py-aiger-dfa)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger-dfa/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger-dfa)\n[![PyPI version](https://badge.fury.io/py/py-aiger-dfa.svg)](https://badge.fury.io/py/py-aiger-dfa)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n<!-- markdown-toc start - Don't edit this section. Run M-x markdown-toc-generate-toc again -->\n**Table of Contents**\n\n- [Installation](#installation)\n- [Usage](#usage)\n    - [DFA to AIG](#dfa-to-aig)\n    - [AIG to DFA](#aig-to-dfa)\n\n<!-- markdown-toc end -->\n\n\n# Installation\n\nIf you just need to use `aiger_dfa`, you can just run:\n\n`$ pip install py-aiger-dfa`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Usage\n\nThe main entry points for using this library are the `dfa2aig` and\n`aig2dfa` functions. DFAs are represented using the\n[dfa](https://github.com/mvcisback/dfa) package. Familiarity with the\n`dfa`, `py-aiger`, and `py-aiger-bv` packages is assumed.\n\n\n## DFA to AIG\n\nAn example of going from a `DFA` to an `AIG` object\nis shown below.\n\n```python\nfrom dfa import DFA\nfrom aiger_dfa import dfa2aig\n\nmy_dfa = DFA(\n    start=0,\n    inputs={0, 1},\n    label=lambda s: (s % 4) == 3,\n    transition=lambda s, c: (s + c) % 4,\n)\nmy_aig, relabels, valid = dfa2aig(my_dfa)\n```\n\nNow `circ` is an `AIG` and `relabels` is a mapping from the inputs,\nstates, and outputs of `my_dfa` to their **1-hot** encoded\ncounterparts in `my_aig`.\n\n`relabels` has the following schema:\n\n```python\nrelabels = {\n    'inputs': .. , #  Mapping from input alphabet -> py-aiger input.\n    'outputs': .. , # Mapping from py-aiger output -> output alphabet.\n    'states': .. , # Mapping from state space -> py-aiger latches.\n}\n```\n\nFinally, `valid` is another aiger circuit which tests if all inputs\nare 1-hot encoded.\n\n## AIG to DFA\n\nWe also support converting a sequential circuit (AIG) to a [Moore\nMachine](https://en.wikipedia.org/wiki/Moore_machine) (DFA) using\n`aig2dfa`. Using the same example:\n\n```python\nfrom aiger_dfa import aig2dfa\n\nmy_dfa2 = aig2dfa(my_aig, relabels=relabels)\n```\n\nNote that the output of a sequential circuit (AIG) is dependent on the\nstate **AND** the action (a [Mealy\nMachine](https://en.wikipedia.org/wiki/Mealy_machine)). \n\nWe use the standard Mealy ↦ Moore reduction where one introduces a\n1-step delay on the output. The default initial output is `None`, but\ncan be set using the `initial_label` argument.\n",
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'mvc@linux.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger-dfa',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



__all__ = ["dfa2aig"]

import aiger_bv
import aiger_bv as BV
import aiger_ptltl as PLTL
from bidict import bidict
from dfa import dfa2dict, DFA
from pyrsistent import pmap

from aiger_dfa.utils import onehot


def dfa2aig(dfa: DFA):
    """
    Takes a dfa.DFA object and returns a tuple of:

    1. An aiger_bv.AIGBV circuit modeling the DFA's labeling and
       transitions.  This circuit has 1 input, output, and latch
       called "action", "output", and "state" respectively.

    2. A dictionary with at three entries, "inputs", "outputs", and
    "states".

      - Each entry is a nested dictionary.
      - The first dictionary maps input/output/state names to encodings.
      - The encoding dictionary is mapping to/from dfa inputs,
        outputs, states and the corresponding aiger inputs, outputs,
        latches.

    3. An aiger_bv.AIGBV circuit which monitors is all inputs are
       valid, e.g., one_hot encoded.
    """
    in2bv, out2bv, state2bv = create_bv_maps(dfa)
    dfa_dict, _ = dfa2dict(dfa)

    action = BV.uatom(len(dfa.inputs), 'action')
    state = BV.uatom(len(dfa.states()), 'state')

    circ = out_circ(dfa_dict, state2bv, out2bv, state)
    circ |= transition_circ(dfa_dict, state2bv, in2bv, action, state)

    start = state2bv[dfa.start]

    circ = circ.loopback({
        "input": "state", "output": "next_state",
        "init": start, "keep_output": False,
    })

    relabels = {
        'inputs': {k: pmap({'action': v}) for k, v in in2bv.items()},
        'outputs': {pmap({'output': v}): k for k, v in out2bv.items()},
        'states': {k: pmap({'state': v}) for k, v in state2bv.items()},
    }

    return circ, relabels, valid_circ(action)


def out_circ(dfa_dict, s2bv, o2bv, state):
    state2label = {s2bv[s]: o2bv[l] for s, (l, _) in dfa_dict.items()}
    tbl = BV.lookup(
        input=state.output,
        inlen=state.size,
        in_signed=False,

        output="output",
        outlen=len(o2bv),            # One hot encoding.
        out_signed=False,

        mapping=state2label,
    )
    return state.aigbv >> tbl


def transition_circ(dfa_dict, s2bv, a2bv, action, state):
    state_action = state.concat(action)

    flattened = {}
    for start, (_, mapping) in dfa_dict.items():
        start = s2bv[start]
        for action, end in mapping.items():
            end = s2bv[end]
            action = a2bv[action]

            flattened[start + action] = end

    tbl = BV.lookup(
        input=state_action.output,
        inlen=state_action.size,
        in_signed=False,

        output="next_state",
        outlen=state.size,             # output: next state
        out_signed=False,

        mapping=flattened,
    )
    return state_action.aigbv >> tbl


def is_1hot(x):
    x = aiger_bv.SignedBVExpr(x.aigbv)
    test = (x != 0) & ((x & (x - 1)) == 0)
    return test.aigbv['o', {test.output: 'valid'}]


def valid_circ(action):
    hist_valid = PLTL.atom('valid').historically().with_output('valid')
    hist_valid = aiger_bv.aig2aigbv(hist_valid.aig)

    return is_1hot(action) >> hist_valid


def _sym2bv(syms):
    size = len(syms)
    return bidict({sym: onehot(i, size) for i, sym in enumerate(syms)})


def create_bv_maps(dfa):
    return map(_sym2bv, [dfa.inputs, dfa.outputs, dfa.states()])
